import cmdHelper from '../../helper/CmdHelper';
import {forEach} from 'min-dash';
import UpdateBusinessObjectHandler from '../../cmd/UpdateBusinessObjectHandler';
import UpdateBusinessObjectListHandler from '../../cmd/UpdateBusinessObjectListHandler';
import {getBusinessObject, is} from "bpmn-js/lib/util/ModelUtil";
import find from 'lodash/find';
import filter from 'lodash/filter';

var HANDLERS = {
    'bpm-modeler.update-businessobject': UpdateBusinessObjectHandler,
    'bpm-modeler.update-businessobject-list': UpdateBusinessObjectListHandler,
};

export default function BpmModeling(eventBus, elementRegistry, moddle, commandStack, bpmnFactory) {
    this._eventBus = eventBus;
    this._elementRegistry = elementRegistry;
    this._moddle = moddle;
    this._commandStack = commandStack;
    this._bpmnFactory = bpmnFactory;

    eventBus.on('diagram.init', function () {
        forEach(HANDLERS, function (handler, id) {
            commandStack.registerHandler(id, handler);
        });
    });
}

BpmModeling.$inject = ['eventBus', 'elementRegistry', 'moddle', 'commandStack', 'bpmnFactory'];

BpmModeling.prototype.updateElementProperties = function (elementId, elementProperties) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }
    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const commands = [];
    commands.push(cmdHelper.updateProperties(element, elementProperties));
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.updateParticipantProcessProperties = function (participantElementId, processProperties) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;

    const participantElement = elementRegistry.get(participantElementId);
    if (!participantElement) {
        return console.error('element not found: ' + participantElementId);
    }
    const bo = participantElement.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + participantElementId);
    }

    const process = bo.get('processRef');
    if (!process) {
        return console.error('processRef for participant is not defined: ' + participantElementId);
    }

    const commands = [];
    commands.push(cmdHelper.updateBusinessObject(participantElement, process, processProperties));
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.removeParticipantProcessExtensionElements = function (participantElementId, extensionElementType) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const commands = [];

    const participantElement = elementRegistry.get(participantElementId);
    if (!participantElement) {
        return console.error('element not found: ' + participantElementId);
    }
    const bo = participantElement.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + participantElementId);
    }
    const process = bo.get('processRef');
    if (!process) {
        return console.error('processRef for participant is not defined: ' + participantElementId);
    }
    const extensionElements = process.extensionElements;
    if (!extensionElements) {
        return;
    }
    var extensionElementsToRemove = filter(extensionElements.get('values'), function (value) {
        return is(value, extensionElementType);
    });
    if (extensionElementsToRemove) {
        commands.push(cmdHelper.removeElementsFromList(process, extensionElements, 'values', null, extensionElementsToRemove));
    }
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createOrUpdateParticipantProcessExtensionElement = function (context) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const participantElementId = context.elementId,
        extensionElementType = context.extensionElementType,
        extensionElementProperties = context.extensionElementProperties,
        alwaysCreateNewElement = context.alwaysCreateNewElement;

    const commands = [];

    const participantElement = elementRegistry.get(participantElementId);
    if (!participantElement) {
        return console.error('element not found: ' + elementId);
    }
    const bo = participantElement.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }
    const process = bo.get('processRef');
    if (!process) {
        return console.error('processRef for participant is not defined: ' + participantElementId);
    }
    let extensionElements = process.extensionElements;
    if (!extensionElements) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, process, bpmnFactory);
        commands.push(cmdHelper.updateBusinessObject(participantElement, process, {extensionElements: extensionElements}));
    }

    let extensionElement;
    if (!alwaysCreateNewElement) {
        const values = extensionElements.get('values');
        for (let i = 0; i < values.length; i++) {
            if (is(values[i], extensionElementType)) {
                extensionElement = values[i];
                break;
            }
        }

        if (extensionElement) {
            commands.push(cmdHelper.removeElementsFromList(process, extensionElements, 'values', null, [extensionElement]));
        }
    }

    extensionElement = createElement(extensionElementType, extensionElementProperties, extensionElements, bpmnFactory);
    commands.push(cmdHelper.addElementsToList(process, extensionElements, 'values', [extensionElement]));

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createOrUpdateNestedObject = function (elementId, nestedPropertyName, nestedElementType, objectProperties) {
    const elementRegistry = this._elementRegistry;
    const bpmnFactory = this._bpmnFactory;
    const commandStack = this._commandStack;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let nestedElement = bo[nestedPropertyName];

    if (!nestedElement) {
        nestedElement = createElement(
            nestedElementType,
            objectProperties,
            bo,
            bpmnFactory
        );
        const propsObj = {};
        propsObj[nestedPropertyName] = nestedElement;
        commands.push(cmdHelper.updateProperties(element, propsObj));
    } else {
        commands.push(cmdHelper.updateBusinessObject(element, nestedElement, objectProperties));
    }

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.setMultiInstanceFormalExpression = function (elementId, propertyName, propertyValue) {
    const elementRegistry = this._elementRegistry;
    const bpmnFactory = this._bpmnFactory;
    const commandStack = this._commandStack;
    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }
    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let loopCharacteristics = bo.get('loopCharacteristics');
    if (!loopCharacteristics) {
        loopCharacteristics = createElement(
            'bpmn:MultiInstanceLoopCharacteristics',
            {},
            bo,
            bpmnFactory
        );
        commands.push(cmdHelper.updateProperties(element, {'multiInstanceLoopCharacteristics': loopCharacteristics}));
    }

    const expressionProps = {};
    if (propertyValue) {
        expressionProps[propertyName] = createElement('bpmn:FormalExpression', {body: propertyValue}, loopCharacteristics, bpmnFactory);
    } else {
        expressionProps[propertyName] = undefined;
    }
    commands.push(cmdHelper.updateBusinessObject(element, loopCharacteristics, expressionProps));

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createOrUpdateFormData = function (elementId, formDataDTO) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let extensionElements = bo.extensionElements;
    if (!extensionElements) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, bo, bpmnFactory);
        commands.push(cmdHelper.updateProperties(element, {extensionElements: extensionElements}));
    }

    let formData;
    const values = extensionElements.get('values');
    for (let i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:FormData')) {
            formData = values[i];
            break;
        }
    }

    if (formData) {
        commands.push(cmdHelper.removeElementsFromList(element, extensionElements, 'values', null, [formData]));
        // var command = cmdHelper.removeElementsFromList(element, extensionElements, 'values', [formData])
        // commandStack.execute(command.cmd, command.context);
    }

    const formDataProperties = {};
    if (formDataDTO.type) formDataProperties.type = formDataDTO.type;
    if (formDataDTO.screenId) formDataProperties.screenId = formDataDTO.screenId;
    if (formDataDTO.openMode) formDataProperties.openMode = formDataDTO.openMode;
    if (formDataDTO.businessKey) formDataProperties.businessKey = formDataDTO.businessKey;

    formData = createElement('jmix:FormData', formDataProperties, extensionElements, bpmnFactory);
    commands.push(cmdHelper.addElementsToList(element, extensionElements, 'values', [formData]));

    if (formDataDTO.fields && formDataDTO.fields.length > 0) {
        const formFields = createElement("jmix:FormFields", null, formData, bpmnFactory);
        commands.push(cmdHelper.updateBusinessObject(element, formData, {formFields: formFields}));
        forEach(formDataDTO.fields, function (formFieldDTO) {
            const formFieldSimpleProperties = {
                'id': formFieldDTO.id,
                'type': formFieldDTO.type,
                'caption': formFieldDTO.caption,
                'editable': formFieldDTO.editable,
                'required': formFieldDTO.required
            };
            const formField = createElement("jmix:FormField", formFieldSimpleProperties, formFields, bpmnFactory);
            commands.push(cmdHelper.addElementsToList(formField, formFields, 'values', [formField]));
            forEach(formFieldDTO.properties, function (formFieldPropertyDTO) {
                const formFieldProperty = createElement("jmix:FormFieldProperty", formFieldPropertyDTO, formField, bpmnFactory);
                commands.push(cmdHelper.addElementsToList(formFieldProperty, formField, 'properties', [formFieldProperty]));
            });
            forEach(formFieldDTO.enumValues, function (formFieldEnumValueDTO) {
                const formFieldEnumValue = createElement("jmix:FormFieldEnumValue", formFieldEnumValueDTO, formField, bpmnFactory);
                commands.push(cmdHelper.addElementsToList(formFieldEnumValue, formField, 'enumValues', [formFieldEnumValue]));
            });
        });
    }

    if (formDataDTO.formParams && formDataDTO.formParams.length > 0) {
        const formParams = createElement("jmix:FormParams", null, formData, bpmnFactory);
        commands.push(cmdHelper.updateBusinessObject(element, formData, {formParams: formParams}));
        forEach(formDataDTO.formParams, function (formParamDTO) {
            const formParam = createElement("jmix:FormParam",
                {
                    'name': formParamDTO.name,
                    'value': formParamDTO.value,
                    'valueSource': formParamDTO.valueSource,
                },
                formData,
                bpmnFactory);
            commands.push(cmdHelper.addElementsToList(element, formParams, 'values', [formParam]));
        });
    }

    if (formDataDTO.outcomes && formDataDTO.outcomes.length > 0) {
        const formOutcomes = createElement("jmix:FormOutcomes", null, formData, bpmnFactory);
        commands.push(cmdHelper.updateBusinessObject(element, formData, {formOutcomes: formOutcomes}));
        forEach(formDataDTO.outcomes, function (formOutcomeDTO) {
            const formOutcome = createElement("jmix:FormOutcome",
                {'id': formOutcomeDTO.id, 'icon': formOutcomeDTO.icon, 'caption': formOutcomeDTO.caption},
                formData,
                bpmnFactory);

            forEach(formOutcomeDTO.outcomeParams, function (paramDTO) {
                const param = createElement("jmix:OutcomeParam",
                    {'name': paramDTO.name, 'value': paramDTO.value},
                    formData,
                    bpmnFactory);
                commands.push(cmdHelper.addElementsToList(element, formOutcome, 'outcomeParams', [param]));
            });

            commands.push(cmdHelper.addElementsToList(element, formOutcomes, 'values', [formOutcome]));
        });
    }

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createOrUpdateProcessVariable = function (elementId, variableName, processVariableDTO) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }
    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let extensionElements = bo.extensionElements;
    if (!extensionElements) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, bo, bpmnFactory);
        commands.push(cmdHelper.updateProperties(element, {extensionElements: extensionElements}));
    }

    let processVariables;
    const values = extensionElements.get('values');
    for (var i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:ProcessVariables')) {
            processVariables = values[i];
            break;
        }
    }

    if (!processVariables) {
        processVariables = createElement('jmix:ProcessVariables', {}, extensionElements, bpmnFactory);
        commands.push(cmdHelper.addElementsToList(element, extensionElements, 'values', [processVariables]));
    }

    let processVariable = find(processVariables.values, function (value) {
        return value.name === variableName;
    });

    if (processVariable) {
        commands.push(cmdHelper.removeElementsFromList(processVariable, processVariables, 'values', null, [processVariable]));
    }

    processVariable = createElement("jmix:ProcessVariable", {name: processVariableDTO.name, type: processVariableDTO.type}, processVariables, bpmnFactory);
    commands.push(cmdHelper.addElementsToList(processVariable, processVariables, 'values', [processVariable]));

    forEach(processVariableDTO.properties, function (processVariablePropertyDTO) {
        const processVariableProperty = createElement("jmix:ProcessVariableProperty", processVariablePropertyDTO, processVariable, bpmnFactory);
        commands.push(cmdHelper.addElementsToList(processVariableProperty, processVariable, 'properties', [processVariableProperty]));
    });

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.removeProcessVariable = function (context) {
    const bo = element.businessObject;
    const elementId = context.elementId,
        variableName = context.variableName;
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;

    const commands = [];

    if (elementId) {
        const element = elementRegistry.get(elementId);
        if (!element) {
            return console.error('element not found: ' + elementId);
        }
    } else {
        return console.error('elementId is defined');
    }

    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const extensionElements = bo.extensionElements;
    if (!extensionElements) {
        return;
    }

    let processVariables;
    const values = extensionElements.get('values');
    for (let i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:ProcessVariables')) {
            processVariables = values[i];
            break;
        }
    }

    if (!processVariables) {
        return;
    }

    const processVariable = find(processVariables.values, function (value) {
        return value.name === variableName;
    });

    if (processVariable) {
        commands.push(cmdHelper.removeElementsFromList(processVariable, processVariables, 'values', null, [processVariable]));
        forEach(commands, function (command) {
            commandStack.execute(command.cmd, command.context);
        });
    }
};

BpmModeling.prototype.createOrUpdateExtensionProperty = function (context) {
    const elementId = context.elementId,
        propertyName = context.propertyName,
        properties = context.properties;
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const commands = [];

    let bo;
    let element;
    if (elementId) {
        element = elementRegistry.get(elementId);
        if (!element) {
            return console.error('element not found: ' + elementId);
        }
        bo = element.businessObject;
        if (!bo) {
            return console.error('businessObject for element is not defined: ' + elementId);
        }
    } else {
        return console.error('elementId is defined');
    }

    let extensionElements = bo.extensionElements;
    if (!extensionElements || !element) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, bo, bpmnFactory);
        commands.push(cmdHelper.updateProperties(element, {extensionElements: extensionElements}));
    }

    let extensionProperties;
    const values = extensionElements.get('values');
    for (let i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:Properties')) {
            extensionProperties = values[i];
            break;
        }
    }

    if (!extensionProperties) {
        extensionProperties = createElement('jmix:Properties', {}, extensionElements, bpmnFactory);
        commands.push(cmdHelper.addElementsToList(element, extensionElements, 'values', [extensionProperties]));
    }

    let extensionProperty = find(extensionProperties.values, function (value) {
        return value.name === propertyName;
    });

    if (!extensionProperty) {
        extensionProperty = createElement("jmix:Property", properties, extensionProperties, bpmnFactory);
        commands.push(cmdHelper.addElementsToList(extensionProperty, extensionProperties, 'values', [extensionProperty]));
    } else {
        for (const pName in properties) {
            extensionProperty[pName] = properties[pName];
        }
    }

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.removeExtensionProperty = function (context) {
    const bo = element.businessObject;
    const elementId = context.elementId,
        propertyName = context.propertyName;
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;

    const commands = [];

    if (elementId) {
        const element = elementRegistry.get(elementId);
        if (!element) {
            return console.error('element not found: ' + elementId);
        }
    } else {
        return console.error('elementId is defined');
    }

    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const extensionElements = bo.extensionElements;
    if (!extensionElements) {
        return;
    }

    let extensionProperties;
    const values = extensionElements.get('values');
    for (let i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:Properties')) {
            extensionProperties = values[i];
            break;
        }
    }

    if (!extensionProperties) {
        return;
    }

    const extensionProperty = find(extensionProperties.values, function (value) {
        return value.name === propertyName;
    });

    if (propertyName) {
        commands.push(cmdHelper.removeElementsFromList(extensionProperty, extensionProperties, 'values', null, [extensionProperty]));
        forEach(commands, function (command) {
            commandStack.execute(command.cmd, command.context);
        });
    }
};

BpmModeling.prototype.createOrUpdateSpringBean = function (context) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;
    const elementId = context.elementId,
        springBeanDTO = context.springBeanDTO;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let extensionElements = bo.extensionElements;
    if (!extensionElements) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, bo, bpmnFactory);
        commands.push(cmdHelper.updateProperties(element, {extensionElements: extensionElements}));
    }

    let springBean;
    const values = extensionElements.get('values');
    for (let i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:SpringBean') && values[i].beanName === springBeanDTO.beanName) {
            springBean = values[i];
            break;
        }
    }

    if (springBean) {
        commands.push(cmdHelper.removeElementsFromList(element, extensionElements, 'values', null, [springBean]));
    }

    const springBeanProperties = {
        'beanName': springBeanDTO.beanName,
        'methodName': springBeanDTO.methodName
    };

    springBean = createElement('jmix:SpringBean', springBeanProperties, extensionElements, bpmnFactory);
    commands.push(cmdHelper.addElementsToList(element, extensionElements, 'values', [springBean]));

    forEach(springBeanDTO.methodParams, function (methodParamDTO) {
        const methodParam = createElement("jmix:MethodParam",
            {
                'name': methodParamDTO.name,
                'type': methodParamDTO.type,
                'value': methodParamDTO.value,
                'isVariable': methodParamDTO.isVariable
            },
            springBean,
            bpmnFactory);
        commands.push(cmdHelper.addElementsToList(element, springBean, 'methodParams', [methodParam]));
    });

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.removeSpringBean = function (context) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const elementId = context.elementId,
        beanName = context.beanName;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const extensionElements = bo.extensionElements;
    if (!extensionElements) {
        return;
    }

    let springBean;
    const values = extensionElements.get('values');
    for (let i = 0; i < values.length; i++) {
        if (is(values[i], 'jmix:SpringBean') && values[i].beanName === beanName) {
            springBean = values[i];
            break;
        }
    }

    if (springBean) {
        commands.push(cmdHelper.removeElementsFromList(element, extensionElements, 'values', null, [springBean]));
    }

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

/**
 * Creates a new extension element with a given type in the element with given id or updates an element with given
 * type if it doesn't exist yet.
 */
BpmModeling.prototype.createOrUpdateExtensionElement = function (context) {
    const elementId = context.elementId,
        extensionElementType = context.extensionElementType,
        extensionElementProperties = context.extensionElementProperties,
        alwaysCreateNewElement = context.alwaysCreateNewElement;

    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let extensionElements = bo.extensionElements;
    if (!extensionElements) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, bo, bpmnFactory);
        commands.push(cmdHelper.updateProperties(element, {extensionElements: extensionElements}));
    }

    let extensionElement;
    if (!alwaysCreateNewElement) {
        const values = extensionElements.get('values');
        for (let i = 0; i < values.length; i++) {
            if (is(values[i], extensionElementType)) {
                extensionElement = values[i];
                break;
            }
        }

        if (extensionElement) {
            commands.push(cmdHelper.removeElementsFromList(element, extensionElements, 'values', null, [extensionElement]));
        }
    }

    extensionElement = createElement(extensionElementType, extensionElementProperties, extensionElements, bpmnFactory);
    commands.push(cmdHelper.addElementsToList(element, extensionElements, 'values', [extensionElement]));

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.removeExtensionElements = function (context) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const elementId = context.elementId,
        extensionElementType = context.extensionElementType;
    const commands = [];
    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const extensionElements = bo.extensionElements;
    if (!extensionElements) {
        return;
    }

    const extensionElementsToRemove = filter(extensionElements.get('values'), function (value) {
        return is(value, extensionElementType);
    });
    if (extensionElementsToRemove) {
        commands.push(cmdHelper.removeElementsFromList(element, extensionElements, 'values', null, extensionElementsToRemove));
    }


    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createListener = function (context) {
    const elementId = context.elementId,
        listenerElementType = context.listenerElementType,
        listenerElementProperties = context.listenerElementProperties,
        fields = context.fields;

    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    let extensionElements = bo.extensionElements;
    if (!extensionElements) {
        extensionElements = createElement('bpmn:ExtensionElements', {values: []}, bo, bpmnFactory);
        commands.push(cmdHelper.updateProperties(element, {extensionElements: extensionElements}));
    }

    let listenerElement;
    listenerElement = createElement(listenerElementType, listenerElementProperties, extensionElements, bpmnFactory);
    listenerElement.fields = [];
    commands.push(cmdHelper.addElementsToList(element, extensionElements, 'values', [listenerElement]));

    forEach(fields, function (field) {
        const fieldElement = createElement("flowable:Field", field, listenerElement, bpmnFactory);
        commands.push(cmdHelper.addElementsToList(element, listenerElement, 'fields', [fieldElement]));
    });

    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.updateTimerEventDefinition = function (elementId, timerDefinitionType, value) {
    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;

    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const timerEventDefinition = getEventDefinition(element, 'bpmn:TimerEventDefinition');
    if (!timerEventDefinition) {
        return console.error('TimerEventDefinition for element is not defined: ' + elementId);
    }
    const props = {
        timeDuration: undefined,
        timeDate: undefined,
        timeCycle: undefined
    };
    props[timerDefinitionType] = createElement('bpmn:FormalExpression', {body: value}, timerEventDefinition, bpmnFactory);
    commands.push(cmdHelper.updateBusinessObject(element, timerEventDefinition, props));
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.updateEventDefinition = function (context) {
    const elementId = context.elementId,
        eventDefinitionType = context.eventDefinitionType,
        referencedElementType = context.referencedElementType,
        referencedElementId = context.referencedElementId,
        propertyName = context.propertyName;

    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }
    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const eventDefinition = getEventDefinition(element, eventDefinitionType);
    if (!eventDefinition) {
        return console.error('Event definition of type ' + eventDefinitionType + ' for element is not defined: ' + elementId);
    }

    const root = getRoot(bo);
    const referencedElement = find(root.rootElements, function (obj) {
        return is(obj, referencedElementType) && obj.id === referencedElementId;
    });

    const props = {};
    props[propertyName] = referencedElement;
    commands.push(cmdHelper.updateBusinessObject(element, eventDefinition, props));
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.updateEventDefinitionProperties = function (context) {
    const elementId = context.elementId,
        eventDefinitionType = context.eventDefinitionType,
        properties = context.properties;

    const elementRegistry = this._elementRegistry;
    const commandStack = this._commandStack;
    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }
    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const eventDefinition = getEventDefinition(element, eventDefinitionType);
    if (!eventDefinition) {
        return console.error('Event definition of type ' + eventDefinitionType + ' for element is not defined: ' + elementId);
    }

    const props = {};
    for (const propertyName in properties) {
        props[propertyName] = properties[propertyName];
    }
    commands.push(cmdHelper.updateBusinessObject(element, eventDefinition, props));
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createOrUpdateRootElement = function (root, elementId, elementType, properties) {
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;
    const commands = [];

    const rootElements = root.rootElements;
    let element = find(rootElements, function (obj) {
        return is(obj, elementType) && obj.id === elementId;
    });
    if (element) {
        commands.push(cmdHelper.updateModdleProperties(element, properties));
    } else {
        element = createElement(elementType, properties, root, bpmnFactory);
        commands.push(cmdHelper.addElementsToList(root, root, 'rootElements', [element]));
    }
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.removeRootElement = function (root, elementId, elementType) {
    const commandStack = this._commandStack;
    const commands = [];

    const rootElements = root.rootElements;
    const element = find(rootElements, function (obj) {
        return is(obj, elementType) && obj.id === elementId;
    });
    if (element) {
        commands.push(cmdHelper.removeElementsFromList(root, root, 'rootElements', null, [element]));
    }
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.createAndSetListProperty = function (context) {
    const elementId = context.elementId,
        listPropertyName = context.listPropertyName,
        elementType = context.elementType,
        objects = context.objects;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;
    const elementRegistry = this._elementRegistry;
    const commands = [];

    const element = elementRegistry.get(elementId);
    if (!element) {
        return console.error('element not found: ' + elementId);
    }

    const bo = element.businessObject;
    if (!bo) {
        return console.error('businessObject for element is not defined: ' + elementId);
    }

    const newObjectList = [];
    if (objects) {
        for (let i = 0; i < objects.length; i++) {
            const newObject = createElement(elementType, objects[i], element, bpmnFactory);
            newObjectList.push(newObject);
        }
    }
    commands.push(cmdHelper.setList(element, bo, listPropertyName, newObjectList));
    forEach(commands, function (command) {
        commandStack.execute(command.cmd, command.context);
    });
};

BpmModeling.prototype.updateProcessDocumentation = function (context) {
    let bo;
    let element;
    const processElementId = context.processElementId,
        participantElementId = context.participantElementId,
        documentationText = context.documentationText;
    const commandStack = this._commandStack;
    const bpmnFactory = this._bpmnFactory;
    const elementRegistry = this._elementRegistry;

    if (processElementId) {
        element = elementRegistry.get(processElementId);
        bo = getBusinessObject(element);
    } else if (participantElementId) {
        element = elementRegistry.get(participantElementId);
        bo = getBusinessObject(element).get('processRef');
    }
    const documentationObject = createElement('bpmn:Documentation', {text: documentationText}, element, bpmnFactory);
    const newObjectList = [documentationObject];
    const command = cmdHelper.setList(element, bo, 'documentation', newObjectList);
    commandStack.execute(command.cmd, command.context);
};

function getEventDefinition(element, eventType) {
    let bo = element.businessObject,
        eventDefinition = null;
    if (bo.eventDefinitions) {
        forEach(bo.eventDefinitions, function (event) {
            if (is(event, eventType)) {
                eventDefinition = event;
            }
        });
    }
    return eventDefinition;
}

BpmModeling.prototype.createElement = createElement;

function createElement(elementType, properties, parent, factory) {
    let element = factory.create(elementType, properties);
    element.$parent = parent;
    return element;
}

function getRoot(businessObject) {
    let parent = businessObject;
    while (parent.$parent) {
        parent = parent.$parent;
    }
    return parent;
}