import {getBusinessObject, is} from 'bpmn-js/lib/util/ModelUtil';
import {filter, find} from 'min-dash';
import * as replaceOptions from "bpmn-js/lib/features/replace/ReplaceOptions";
import {isExpanded} from "bpmn-js/lib/util/DiUtil";
import {doPatchGetEntry} from "../../patching/diagramjs-patcher";

/**
 * This module is an element agnostic replace menu provider for the popup menu.
 */
export default function BpmReplaceMenuProvider(popupMenu, replaceMenuProvider, bpmnFactory, bpmnReplace, bpmModeling) {
    doPatchGetEntry(popupMenu);
    this._popupMenu = popupMenu;
    this._replaceMenuProvider = replaceMenuProvider;
    this._bpmnFactory = bpmnFactory;
    this._bpmnReplace = bpmnReplace;
    this._bpmModeling = bpmModeling;
    this.register();
}

BpmReplaceMenuProvider.$inject = [
    'popupMenu',
    'replaceMenuProvider',
    "bpmnFactory",
    "bpmnReplace",
    'bpmModeling',
];


/**
 * Register replace menu provider in the popup menu
 */
BpmReplaceMenuProvider.prototype.register = function () {
    this._popupMenu.registerProvider('bpmn-replace', this);
};


/**
 * Get all entries from replaceOptions for the given element and apply filters
 * on them. Get for example only elements, which are different from the current one.
 *
 * @param {djs.model.Base} element
 *
 * @return {Array<Object>} a list of menu entry items
 */
BpmReplaceMenuProvider.prototype.getPopupMenuEntries = function (element) {
    return function () {
        const businessObject = element.businessObject;
        let entries = is(businessObject, 'bpmn:Task') ?
            _getEntries(element, this._replaceMenuProvider) :
            this._replaceMenuProvider.getEntries(element);

        //remove not supported elements from replace options
        const elementsToExclude = [
            //conditional
            'replace-with-conditional-start',
            'replace-with-non-interrupting-conditional-start',
            'replace-with-conditional-intermediate-catch',
            'replace-with-conditional-boundary',
            'replace-with-non-interrupting-conditional-boundary',
            //task
            'replace-with-receive-task',
            'replace-with-send-task',
            'replace-with-manual-task',
            'replace-with-collapsed-subprocess',
            //escalation
            'replace-with-escalation-start',
            'replace-with-escalation-intermediate-throw',
            'replace-with-escalation-boundary',
            'replace-with-non-interrupting-escalation-boundary',
            'replace-with-escalation-end',
            //link
            'replace-with-link-intermediate-catch',
            'replace-with-link-intermediate-throw',
            //message
            'replace-with-message-intermediate-throw',
            //end events
            'replace-with-message-end',
            'replace-with-compensation-end',
            'replace-with-signal-end',
        ];
        entries = entries.filter(function (entry) {
            return elementsToExclude.indexOf(entry.id) < 0;
        });

        //we get all entries from the standard ReplaceMenuProvider, but the standard 'replace-with-rule-task' must be replaced.
        //Flowable engine uses a ServiceTask with an attribute type='dmn' for DMN tasks. That's why we provide a custom replace action
        if (is(businessObject, 'bpmn:Task')) {
            const bpmnReplace = this._bpmnReplace;
            const bpmModeling = this._bpmModeling;

            const replaceWithRuleTaskEntry = find(entries, function (e) {
                return e.id === 'replace-with-rule-task'
            });
            if (replaceWithRuleTaskEntry) {
                const definition = {
                    label: 'Business Rule Task',
                    actionName: 'replace-with-rule-task',
                    className: 'bpmn-icon-business-rule',
                    target: {
                        type: 'bpmn:BusinessRuleTask'
                    }
                };
                const newReplaceWithRuleTaskEntry = this._replaceMenuProvider._createMenuEntry(definition, element, function () {
                    const newElement = bpmnReplace.replaceElement(element, {type: "bpmn:ServiceTask"});
                    bpmModeling.updateElementProperties(newElement.id, {'type': 'dmn', 'taskType': null});
                    return newElement;
                });
                const index = entries.indexOf(replaceWithRuleTaskEntry);
                entries[index] = newReplaceWithRuleTaskEntry;
            }

            const updateServiceTaskEntry = find(entries, function (e) {
                return e.id === 'replace-with-service-task'
            });

            if (updateServiceTaskEntry) {
                const stDefinition = {
                    label: 'Service Task',
                    actionName: 'replace-with-service-task',
                    className: 'bpmn-icon-service',
                    target: {
                        type: 'bpmn:ServiceTask'
                    }
                };
                const newServiceTaskEntry = this._replaceMenuProvider._createMenuEntry(stDefinition, element, function () {
                    const newElement = bpmnReplace.replaceElement(element, {type: 'bpmn:ServiceTask'});
                    //clear business-rule-task and entity-data-task properties
                    bpmModeling.updateElementProperties(newElement.id, {
                        'type': null,
                        'taskType': null,
                        'delegateExpression': null
                    });
                    bpmModeling.removeExtensionElements({
                        elementId: newElement.id,
                        extensionElementType: 'flowable:Field'
                    });
                    return newElement;
                });
                const stIndex = entries.indexOf(updateServiceTaskEntry);
                entries[stIndex] = newServiceTaskEntry;
            }

            if (!(['jmix-load-entities-jpql', 'jmix-modify-entity', 'jmix-create-entity'].includes(businessObject.type))) {
                const entityDataTaskMenuEntry = this._replaceMenuProvider._createMenuEntry(
                    {
                        label: 'Entity Data Task',
                        actionName: 'replace-with-entity-data-task',
                        className: 'bpmn-icon-data-store',
                        target: {
                            type: 'bpmn:ServiceTask'
                        }
                    },
                    element,
                    function () {
                        const newElement = bpmnReplace.replaceElement(element, {type: "bpmn:ServiceTask"});
                        bpmModeling.updateElementProperties(newElement.id, {
                            'type': 'jmix-load-entities-jpql',
                            'taskType': null,
                            'delegateExpression': null
                        });
                        bpmModeling.removeExtensionElements({
                            elementId: newElement.id,
                            extensionElementType: 'flowable:Field'
                        });
                        return newElement;
                    }
                );
                entries.push(entityDataTaskMenuEntry);
            }

            if ('jmix-send-email' !== businessObject.type) {
                const sendEmailTaskMenuEntry = this._replaceMenuProvider._createMenuEntry(
                    {
                        label: 'Send Email Task',
                        actionName: 'replace-with-send-email-task',
                        className: 'bpmn-icon-send',
                        target: {
                            type: 'bpmn:ServiceTask'
                        }
                    },
                    element,
                    function () {
                        const newElement = bpmnReplace.replaceElement(element, {type: "bpmn:ServiceTask"});
                        bpmModeling.updateElementProperties(newElement.id, {
                            'type': 'jmix-send-email',
                            'taskType': null,
                            'delegateExpression': null
                        });
                        bpmModeling.removeExtensionElements({
                            elementId: newElement.id,
                            extensionElementType: 'flowable:Field'
                        });
                        return newElement;
                    }
                );
                entries.push(sendEmailTaskMenuEntry);
            }
        }
        return entries;
    }.bind(this);
};

/**
 * Hides default buttons for multi instance markers and for the ad hoc marker
 *
 * @param element
 * @returns {any}
 */
BpmReplaceMenuProvider.prototype.getPopupMenuHeaderEntries = function (element) {
    return function () {
        return {};
    }
};

function _getEntries(element, replaceMenuProvider) {
    const differentType = _isDifferentType(element);
    const entries = filter(replaceOptions.TASK, differentType);
    return replaceMenuProvider._createEntries(element, entries);
}

//copied from lib/features/popup-menu/util/TypeUtil.js with modification for BusinessRuleTask and EntityDataTask
function _isDifferentType(element) {
    return function (entry) {
        const target = entry.target;
        const businessObject = getBusinessObject(element),
            eventDefinition = businessObject.eventDefinitions && businessObject.eventDefinitions[0];

        const isTypeEqual = businessObject.$type === target.type;

        //businessRuleTask modeled as serviceTask with type == 'dmn'
        if (businessObject.type === 'dmn') {
            if (isTypeEqual) {
                return true;
            } else if (target.type === 'bpmn:BusinessRuleTask') {
                return false;
            }
        }

        //entityDataTasks and sendEmailTask modeled as serviceTask with appropriate type 'jmix-...'
        if (isTypeEqual
            && ['jmix-load-entities-jpql', 'jmix-modify-entity', 'jmix-create-entity', 'jmix-send-email'].includes(businessObject.type)) {
            return true;
        }

        const isEventDefinitionEqual = (
            (eventDefinition && eventDefinition.$type) === target.eventDefinitionType
        );

        const isTriggeredByEventEqual = (
            businessObject.triggeredByEvent === target.triggeredByEvent
        );

        const isExpandedEqual = (
            target.isExpanded === undefined ||
            target.isExpanded === isExpanded(businessObject));

        return !isTypeEqual || !isEventDefinitionEqual || !isTriggeredByEventEqual || !isExpandedEqual;
    }
}