/*
 * Copyright (c) Haulmont 2023. All Rights Reserved.
 * Use is subject to license terms.
 */

// @ts-ignore
import {html, LitElement, css} from 'lit';
// @ts-ignore
import {customElement} from 'lit/decorators.js';
import BpmViewer from "./bpm/BpmViewer";
import selectionChangedEventHandler from "./component/selectionChangedEventHandler";
import {JmixBpmIncomingEvent} from "./component/types";

@customElement('jmix-bpmn-viewer')
class JmixBpmViewer extends LitElement {
    private readonly BPM_VIEWER_HOLDER: string = "bpmnViewerHolder";
    private readonly viewer: BpmViewer;
    private bpmnXml: string | undefined;
    private readonly bpmDrawing: any;
    private readonly canvas: any;
    private shadowRoot: any;

    static styles = css`
      .highlighted:not(.djs-connection) .djs-visual > :nth-child(1) {
        fill: #c2d5ed !important;
      }
    `;

    constructor() {
        super();
        this.viewer = new BpmViewer();
        this.bpmDrawing = this.viewer.get("bpmDrawing");
        this.canvas = this.viewer.get("canvas");

        let viewer = this.viewer as any;
        const canvas = viewer.get("canvas");
        const element = this;

        viewer.on("selection.changed", e => selectionChangedEventHandler(element, canvas.getRootElement(), e as JmixBpmIncomingEvent));
    }

    static get properties() {
        return {
            bpmnXml: {type: String}
        }
    }

    public async reloadSchema(xmlSchema: string) {
        await this.viewer.importXML(xmlSchema)
    }

    // @ts-ignore
    public async getXmlSchema(): Promise<string> {
        const requestXml = await this.viewer.saveXML({
            format: true
        });
        return requestXml.xml!;
    }

    render() {
        return html`
            <div id="${(this.BPM_VIEWER_HOLDER)}" style="width: 100%; height: 100%"/>
        `;
    }

    updated(updatedProps) {
        this.awaitRun(() => this.initViewer());
    }

    public setElementColor(cmdJson) {
        const cmd = JSON.parse(cmdJson);
        this.awaitRun(() => this.bpmDrawing.setElementColor(cmd));
    }

    public addMarker(cmdJson) {
        const cmd = JSON.parse(cmdJson);
        this.awaitRun(() => this.canvas.addMarker(cmd.elementId, cmd.marker));
    }

    public removeMarker(cmdJson) {
        const cmd = JSON.parse(cmdJson);
        this.awaitRun(() => this.canvas.removeMarker(cmd.elementId, cmd.marker));
    }

    private initViewer() {
        this.viewer.attachTo(this.shadowRoot.getElementById(this.BPM_VIEWER_HOLDER)!);
        if (this.bpmnXml) {
            this.viewer.importXML(this.bpmnXml).then(r => {
                console.log(r.warnings);
            });
        }
    }

    private awaitRun(callable: () => void) {
        setTimeout(() => {
            try {
                callable();
            } catch (err) {
                console.log('error while running callable', err);
            }
        }, 100);
    }
}
